import os
from collections import defaultdict
from typing import TYPE_CHECKING, List

from potion import Request
from potion.api import *
from potion.objects import Parent, Properties, Database, sche
from .base import NotionObject

if TYPE_CHECKING:
    from .page import NotionPage


class NotionDatabase(NotionObject):
    def __init__(self, auth=None, id=None, parent: Parent = None):
        super().__init__(auth, id, parent)
        if auth is None:
            auth = os.environ.get('NOTION_SECRET', None)
            assert auth is not None, 'no auth'

        if id is None:
            assert parent is not None, 'no page id'

        self.req = Request.from_token(authorization=auth)
        self.auth = auth
        if id is None:
            assert parent is not None
            res = self.req.post(database_create(), data=Database(parent=parent, properties=Properties(
                sche.Title(pname='name'),
            )))
            if isinstance(res, Database):
                id = res.id
                parent = res.parent
                self.id = id
            else:
                assert False, res.to_json(2)

        else:
            res = self.req.get(database_retrieve(id))
            assert isinstance(res, Database), res.to_json(2)
            parent = res.parent
            self.id = id
        self.object = res
        self.blocks = {}

        self._parent = parent
        self.cache_properties = []
        self.children = []
        self.multi_select_property = defaultdict(set)

    def create_page(self) -> 'NotionPage':
        from .page import NotionPage
        return NotionPage(auth=self.auth, parent=Parent.DataBaseParent(self.id))

    def properties(self):
        return self.object.properties

    def set_property(self, property_args: sche.Schema):
        self.cache_properties.append(property_args)

    def flush_property(self):
        return self.update_properies(self.cache_properties)

    def update_properies(self, property_list: List[sche.Schema]):
        p = Properties(*property_list)
        presp = self.req.patch(database_update(self.id), Database(properties=Properties(p)))
        if isinstance(presp, Database):
            self.object = presp
            return True, presp
        else:
            return False, presp



default_headers = {
    'Content-Type': 'application/json',
    'Notion-Version': '2022-02-22'
}

current_headers = {}


class NotionHeader:
    def __init__(self, authorization, notion_version=None):
        self._headers = {
            'Authorization': authorization
        }
        if notion_version is not None:
            self._headers['Notion-Version'] = notion_version

    @property
    def headers(self) -> dict:
        val = {}
        val.update(default_headers)
        val.update(self._headers)
        return val

    # @contextmanager
    # def __call__(self):
    #     headers = {}
    #     headers.update(default_headers)
    #     headers['Authorization'] = self.authorization
    #     current_headers.update(headers)
    #     if self.notion_version is not None:
    #         headers['Notion-Version'] = self.notion_version
    #     try:
    #         yield headers
    #     finally:
    #         current_headers.clear()



from typing import List

__all__ = ['Number', 'Select', 'MultiSelect', 'Date', 'Option',
           'CheckBox', 'URL', 'Email', 'Phone', 'Title', 'RichText']

from .common import NotionObject, Any, Null


class Schema(NotionObject): pass


class AnySchema(Any): pass


class DeleteSchema(Schema):
    def __init__(self, pname: str):
        super().__init__(pname, args=Null)


class Title(Schema):
    def __init__(self, pname: str):
        super().__init__(pname)


class RichText(Schema):
    property_name = 'rich_text'

    def __init__(self, pname: str):
        super().__init__(pname)


class Number(Schema):
    def __init__(self, pname: str, format: str):
        super().__init__(pname, kwargs=dict(format=format))


class Option(Schema):
    def __init__(self, name, color=None):
        super().__init__(None, kwargs=dict(name=name, color=color))

    def to_dict(self):
        val = {
            'name': self['name'],
        }
        if self['color'] is not None:
            val['color'] = self['color']
        return val


class Options(Schema):
    def __init__(self, options: List[Option]):
        super().__init__('', args=options)


class Select(Schema):
    def __init__(self, pname, options: List[Option]):
        super().__init__(pname, args=Options(options))


class MultiSelect(Schema):
    property_name = 'multi_select'
    include_type = True

    def __init__(self, pname, options: List[Option]):
        super().__init__(pname, args=Options(options))


class SingleValueSchema(Schema):
    def __init__(self, pname: str):
        super().__init__(pname)


class Date(SingleValueSchema):
    pass


class People(SingleValueSchema):
    pass


class Files(SingleValueSchema):
    pass


class CheckBox(SingleValueSchema): pass


class URL(SingleValueSchema): pass


class Email(SingleValueSchema): pass


class Phone(SingleValueSchema): pass


class Formula(Schema):
    def __init__(self, pname: str):
        super().__init__(pname)
        raise NotImplementedError()


class Relation(Schema):
    def __init__(self, pname: str):
        super().__init__(pname)
        raise NotImplementedError('Relation property is not supported currently.')


class Rollup(Schema):
    def __init__(self, pname: str):
        super().__init__(pname)
        raise NotImplementedError('Rollup property is not supported currently.')


class CreatedTime(SingleValueSchema):
    property_name = 'created_time'


class CreatedBy(SingleValueSchema):
    property_name = 'created_by'


class LastEditedTime(SingleValueSchema):
    property_name = 'last_edited_time'


class LastEditedBy(SingleValueSchema):
    property_name = 'last_edited_by'



from .common import NotionObject


class Condition(NotionObject):
    def __init__(self, *properties):
        super().__init__('', args=properties)


class And(Condition): pass


class Or(Condition): pass


class SortProperty(NotionObject):
    def __init__(self, property=None, direction='ascending'):
        super().__init__(None, kwargs={'property': property, 'direction': direction})


class QueryProperty(NotionObject):
    """
    see https://developers.notion.com/reference/post-database-query-filter for details
    """
    union_type = 'dict'

    def __init__(self, property_name: str, property_type: str, conditions: str, condition_value):
        super().__init__(None, kwargs={
            'property': property_name,
            property_type: {
                conditions: condition_value
            }
        })


class Filter(NotionObject):
    union_type = 'dict'

    def __init__(self, pname: str, conditions: Condition):
        super().__init__(pname, args=conditions)

    @staticmethod
    def Search(conditions: Condition):
        return Filter(None, conditions)

    @staticmethod
    def QueryDatabase(filter: Condition):
        return Filter('', filter)


class Sorts(NotionObject):
    property_name = 'sorts'

    def __init__(self, *sorts):
        super().__init__('', args=sorts)


class Search(NotionObject):

    def __init__(self, query: str = None, sort=None, filter=None, start_cursor: str = None, page_size: int = None):
        super().__init__(None, kwargs={
            'query': query,
            'sort': sort,
            'filter': filter,
            'start_cursor': start_cursor,
            'page_size': page_size,
        })



from .common import NotionObject


class Block(NotionObject):
    union_type = 'dict'

    def __init__(self, **kwargs):
        super().__init__('',
                         kwargs=dict(**kwargs))

    def set_children(self, blocks):
        self['children'] = blocks

    @property
    def id(self):
        return self['id']

    @property
    def has_children(self):
        return self['has_children']

    @property
    def type(self):
        return self['type']


class Todo(Block):
    property_name = 'to_do'

    def __init__(self, rich_text=None, checked=None, color=None, children=None):
        super().__init__(rich_text=rich_text,
                         checked=checked,
                         color=color,
                         children=children)


class Callout(Block):

    def __init__(self, rich_text=None, icon=None, color=None, children=None):
        super().__init__(rich_text=rich_text,
                         icon=icon,
                         color=color,
                         children=children)


class Quote(Block):

    def __init__(self, rich_text=None, color=None, children=None):
        super().__init__(rich_text=rich_text,
                         color=color,
                         children=children)


class Code(Block):

    def __init__(self, rich_text=None, caption=None, language=None):
        super().__init__(rich_text=rich_text,
                         caption=caption,
                         language=language)


class Bulleted(Block):
    property_name = 'bulleted_list_item'

    def __init__(self, rich_text=None, color=None, children=None):
        super().__init__(rich_text=rich_text,
                         color=color,
                         children=children)

    @property
    def rich_text(self):
        return self['rich_text']

    @property
    def color(self):
        return self['color']

    @property
    def children(self):
        return self['children']


class Numbered(Bulleted):
    property_name = 'numbered_list_item'


class Toggle(Bulleted):
    property_name = 'toggle'


class _Headeing(Block):

    def __init__(self, rich_text=None, color=None):
        super().__init__(rich_text=rich_text,
                         color=color)


class Heading1(_Headeing):
    property_name = 'heading_1'


class Heading2(_Headeing):
    property_name = 'heading_2'


class Heading3(_Headeing):
    property_name = 'heading_3'



from potion.objects import *


def parse(dic: dict) -> NotionObject:
    if dic['object'] == 'error':
        return Error(**dic)
    elif dic['object'] == 'page':
        return Page(**dic)
    elif dic['object'] == 'database':
        return Database(**dic)
    elif dic['object'] == 'block':
        return Block(**dic)
    elif dic['object'] == 'list':
        return Pagination(**dic)
    return dic



from .delete import *
from .get import *
from .patch import *
from .post import *

__all__ = [
    'database_retrieve', 'database_update', 'database_create', 'database_query',
    'page_update', 'page_retrieve', 'page_property_retrieve', 'page_create',
    'block_delete', 'block_retrieve', 'block_children_append', 'block_children_retrieve', 'block_update',
    'users_list', 'user_retrieve', 'user_bot_retrieve', 'search'
]



class Color:
    default = 'default'
    gray = "gray"
    brown = "brown"
    orange = "orange"
    yellow = "yellow"
    green = "green"
    blue = "blue"
    purple = "purple"
    pink = "pink"
    red = "red"
    gray_background = "gray_background"
    brown_background = "brown_background"
    orange_background = "orange_background"
    yellow_background = "yellow_background"
    green_background = "green_background"
    blue_background = "blue_background"
    purple_background = "purple_background"
    pink_background = "pink_background"
    red_backgroun = "red_backgroun"


class FilterProperty:
    rich_text = 'rich_text'
    phone_number = 'phone_number'
    number = 'number'
    checkbox = 'checkbox'
    select = 'select'
    multi_select = 'multi-select'
    date = 'date'
    people = 'people'
    files = 'files'
    relation = 'relation'
    formula = 'formula'


class FilterCondition:
    class Text:
        # string
        equals = 'equals'
        does_not_equal = 'does_not_equal'
        contains = 'contains'
        does_not_contain = 'does_not_contain'
        starts_with = 'starts_with'
        ends_with = 'ends_with'
        # boolean (only True)
        is_empty = 'is_empty'
        is_not_empty = 'is_not_empty'

    class Number:
        # string
        equals = 'equals'
        does_not_equal = 'does_not_equal'
        greater_than = 'greater_than'
        less_than = 'less_than'
        greater_than_or_equal_to = 'greater_than_or_equal_to'
        less_than_or_equal_to = 'less_than_or_equal_to'
        # boolean (only True)
        is_empty = 'is_empty'
        is_not_empty = 'is_not_empty'

    class Checkbox:
        # boolean
        equals = 'equals'
        does_not_equal = 'does_not_equal'

    class Select:
        # string
        equals = 'equals'
        does_not_equal = 'does_not_equal'
        # boolean (only True)
        is_empty = 'is_empty'
        is_not_empty = 'is_not_empty'

    class MultiSelecte:
        # string
        contains = 'contains'
        does_not_contain = 'does_not_contain'
        # boolean (only True)
        is_empty = 'is_empty'
        is_not_empty = 'is_not_empty'

    # TODO Formula
    # TODO Rollup
    # TODO Relation
    # TODO Files
    # TODO People
    class Date:
        equals = 'equals'
        is_empty = 'is_empty'
        is_not_empty = 'is_not_empty'
        before = 'before'
        after = 'after'
        on_or_before = 'on_or_before'
        on_or_after = 'on_or_after'
        past_week = 'past_week'
        past_month = 'past_month'
        past_year = 'past_year'
        next_week = 'next_week'
        next_month = 'next_month'
        next_year = 'next_year'


class Language:
    abap = 'abap'
    arduino = 'arduino'
    bash = 'bash'
    basic = 'basic'
    c = 'c'
    clojure = 'clojure'
    coffeescript = 'coffeescript'
    cpp = 'c++'
    csharp = 'c#'
    css = 'css'
    dart = 'dart'
    diff = 'diff'
    docker = 'docker'
    elixir = 'elixir'
    elm = 'elm'
    erlang = 'erlang'
    flow = 'flow'
    fortran = 'fortran'
    fsharp = 'f#'
    gherkin = 'gherkin'
    glsl = 'glsl'
    go = 'go'
    graphql = 'graphql'
    groovy = 'groovy'
    haskell = 'haskell'
    html = 'html'
    java = 'java'
    javascript = 'javascript'
    json = 'json'
    julia = 'julia'
    kotlin = 'kotlin'
    latex = 'latex'
    less = 'less'
    lisp = 'lisp'
    livescript = 'livescript'
    lua = 'lua'
    makefile = 'makefile'
    markdown = 'markdown'
    markup = 'markup'
    matlab = 'matlab'
    mermaid = 'mermaid'
    nix = 'nix'
    objective_c = 'objective-c'
    ocaml = 'ocaml'
    pascal = 'pascal'
    perl = 'perl'
    php = 'php'
    plain_text = 'plain text'
    powershell = 'powershell'
    prolog = 'prolog'
    protobuf = 'protobuf'
    python = 'python'
    r = 'r'
    reason = 'reason'
    ruby = 'ruby'
    rust = 'rust'
    sass = 'sass'
    scala = 'scala'
    scheme = 'scheme'
    scss = 'scss'
    shell = 'shell'
    sql = 'sql'
    swift = 'swift'
    typescript = 'typescript'
    vb_net = 'vb.net'
    verilog = 'verilog'
    vhdl = 'vhdl'
    visual_basic = 'visual basic'
    webassembly = 'webassembly'
    xml = 'xml'
    yaml = 'yaml'
    jcps = 'java/c/c++/c#'



def databases(database_id='', query=False):
    query_str = '/query' if query else ''
    return f'https://api.notion.com/v1/databases/{database_id}{query_str}'


def pages(page_id='', property_id=None):
    property_str = f'/properties/{property_id}' if property_id is not None else ''
    return f'https://api.notion.com/v1/pages/{page_id}{property_str}'


def blocks(block_id='', page_size=None, start_cursor=None, append=False):
    assert page_size is None or isinstance(page_size, int)
    # assert start_cursor is None or isinstance(start_cursor, star)

    if append:
        children_str = f'/children'
    else:
        children_strs = []
        if page_size is not None:
            children_strs.append(f'page_size={page_size}')
        if start_cursor is not None:
            children_strs.append(f'start_cursor={start_cursor}')
        children_strs = '&'.join(children_strs)
        children_str = f'/children?{children_strs}'
    return f"https://api.notion.com/v1/blocks/{block_id}{children_str}"


def blocks_cursor(block_id='', page_size=None, start_cursor=None):
    assert page_size is None or isinstance(page_size, int)
    return f"https://api.notion.com/v1/blocks/{block_id}/children?"


def users(user_id=''):
    return f'https://api.notion.com/v1/users/{user_id}'


def search():
    return 'https://api.notion.com/v1/search'



# b'{"object":"error","status":401,"code":"unauthorized","message":"API token is invalid."}'
from .common import NotionObject


class Error(NotionObject):
    def __init__(self, status: int, code: str, message: str, object=None):
        super().__init__(None, kwargs=dict(status=status, code=code, message=message))



import os
from collections import defaultdict
from datetime import datetime
from typing import Union, TYPE_CHECKING

from potion import Request
from potion.api import *
from potion.objects import Parent, Page, Properties
from potion.objects import block, rich, prop, Error as NotionError
from .base import NotionObject

if TYPE_CHECKING:
    from .database import NotionDatabase


class NotionPage(NotionObject):
    def __init__(self, auth=None, id=None, parent: Parent = None):
        super().__init__(auth, id, parent)
        if auth is None:
            auth = os.environ.get('NOTION_SECRET', None)
            assert auth is not None, 'no auth'

        if id is None:
            assert parent is not None, 'no page id'

        self.req = Request.from_token(authorization=auth)
        self.auth = auth
        if id is None:
            assert parent is not None
            res = self.req.post(page_create(), data=Page(parent=parent, properties=Properties()))
            if isinstance(res, Page):
                id = res.id
                self.id = id
            else:
                assert False, res.to_json(2)
        else:
            res = self.req.get(page_retrieve(id))
            assert isinstance(res, Page), res.to_json(2)
            self.id = id
        self.object = res
        self.blocks = {}

        self._parent = parent

        self.properties = []
        self.children = []
        self.multi_select_property = defaultdict(set)

    def flush_property(self):
        for property_name, values in self.multi_select_property.items():
            self.properties.append(prop.MultiSelect(property_name, selects=[
                prop.MultiSelectOption(name=v) for v in values
            ]))

        page = Page(properties=Properties(*self.properties))
        res = self.req.patch(
            url=page_update(self.id),
            data=page
        )
        self.properties.clear()
        if isinstance(res, NotionError):
            print(res)
        return res

    def flush_children(self):
        page = Page(children=self.children)
        res = self.req.patch(
            url=block_children_append(self.id),
            data=page
        )
        self.children.clear()
        if isinstance(res, NotionError):
            print(res)
        return res

    def append_content(self, content: Union[dict, block.Block]):
        self.children.append(content)

    def append_code(self, code, language='python'):
        code = [code[i:i + 1999] for i in range(0, len(code), 1999)]
        content = block.Code(rich_text=[rich.Text(f"{i}") for i in code][:100], language=language)
        self.append_content(content)

    def create_database(self) -> 'NotionDatabase':
        from .database import NotionDatabase
        self.flush_children()
        return NotionDatabase(auth=self.auth, parent=Parent.PageParent(self.id))

    @property
    def title_property_name(self):
        for prop_name in self.object.properties:
            if self.object.properties[prop_name]['id'] == 'title':
                return prop_name

    def set_title(self, value: str):
        self.properties.append(prop.Title(self.title_property_name,
                                          rich_text=[rich.Text(content=value)]))

    def set_text(self, property_name, value):
        value = f'{value}'
        self.properties.append(prop.RichTextProp(property_name,
                                                 rich_text=[rich.Text(content=value)]))

    def set_duration(self, property_name: str, start: datetime = None, end: datetime = None):
        self.properties.append(prop.Date(property_name, start=start, end=end))

    def set_checkbox(self, property_name: str, checkbox=True):
        self.properties.append(prop.CheckBox(property_name, checkbox))

    def set_number(self, property_name: str, number):
        self.properties.append(prop.Number(property_name, number))

    def add_option(self, property_name, value):
        if value is None:
            return
        else:
            value = f'{value}'
        self.multi_select_property[property_name].add(value)
        self.properties.append(prop.MultiSelect(pname=property_name, selects=[
            prop.MultiSelectOption(name=value)
        ]))

    def remove_option(self, property_name, value):
        if value in self.multi_select_property[property_name]:
            self.multi_select_property[property_name].remove(value)



from datetime import datetime
from typing import List, Union, Dict

__all__ = ['Number', 'Select', 'MultiSelect', 'Date',
           'CheckBox', 'URL', 'Email', 'Phone', 'Property']

from .common import NotionObject, _Null
from .rich import RichText


class Property(NotionObject): pass


class Title(Property):
    def __init__(self, pname: str, rich_text: List[Union[Dict, RichText]]):
        super().__init__(pname, args=rich_text)

    @staticmethod
    def PageTitle(rich_text: List[Union[Dict, RichText]]):
        return Title(None, rich_text=rich_text)


class RichTextProp(Property):
    property_name = 'rich_text'

    def __init__(self, pname: str, rich_text: List[Union[Dict, RichText]]):
        super().__init__(pname, args=rich_text)

    @staticmethod
    def PageTitle(rich_text: List[Union[Dict, RichText]]):
        return Title(None, rich_text=rich_text)


class Number(Property):
    def __init__(self, pname: str, number: float):
        super().__init__(pname, args=number)


class Select(Property):
    def __init__(self, pname, id=None, name=None, color=None):
        super().__init__(pname, kwargs=dict(id=id, name=name, color=color))

    @property
    def id(self):
        return self['id']

    @property
    def name(self):
        return self['name']

    @property
    def color(self):
        return self['color']


class MultiSelectOption(Property):
    def __init__(self, name=None):
        super().__init__(None, kwargs=dict(name=name))


class MultiSelect(Property):
    property_name = 'multi_select'

    def __init__(self, pname, selects: Union[List[MultiSelectOption], _Null]):
        if isinstance(selects, list):
            selects = [i._value for i in selects]
        super().__init__(pname, args=selects)


class Date(Property):
    def __init__(self, pname, start: datetime = None, end: datetime = None, time_zone: str = None):
        if start is not None:
            start = start.isoformat()
        if end is not None:
            end = end.isoformat()
        kwargs = dict(start=start,
                      end=end,
                      time_zone=time_zone)
        super(Date, self).__init__(pname,
                                   kwargs=kwargs)


class SingleValueProperty(Property):
    def __init__(self, pname: str, value=None):
        super().__init__(pname, args=value)


class CheckBox(SingleValueProperty):
    """
    当 check = None 时，输出格式为
    {
      "Done?": {
        "checkbox": {}
      }
    }
    否则为单值
    """

    def __init__(self, pname: str, check: bool = None):
        super().__init__(pname, value=check)


class URL(SingleValueProperty):
    def __init__(self, pname: str, url: str = None):
        super().__init__(pname, value=url)


class Email(SingleValueProperty):
    def __init__(self, pname: str, email: str = None):
        super().__init__(pname, value=email)


class Phone(SingleValueProperty):
    def __init__(self, pname: str, phone_number: str = None):
        super().__init__(pname, value=phone_number)


class Formula(Property):
    def __init__(self, pname: str):
        super().__init__(pname)
        raise NotImplementedError('Formula property is not supported currently.')


class Relation(Property):
    def __init__(self, pname: str):
        super().__init__(pname)
        raise NotImplementedError('Relation property is not supported currently.')


class Rollup(Property):
    def __init__(self, pname: str):
        super().__init__(pname)
        raise NotImplementedError('Rollup property is not supported currently.')


class People(Property):
    def __init__(self, pname: str):
        super().__init__(pname)
        raise NotImplementedError('People property is not supported currently.')


class Files(Property):
    def __init__(self, pname: str):
        super().__init__(pname)
        raise NotImplementedError('Files property is not supported currently.')



import json
from typing import List, Dict, Tuple


class _Null: pass


Null = _Null()


class NotionObject:
    property_name = None
    union_type = 'list'
    include_type = False

    def __init__(self, pname: str, args=None, kwargs=None):
        self._name = pname
        self._loaded_keys = set()
        if args is not None:
            self._value = args
        elif kwargs is not None:
            self._value = {k: v for k, v in kwargs.items() if v is not None}
        else:
            self._value = {}

    def load_mark(self, key):
        self._loaded_keys.add(key)

    def loaded(self, key):
        return key in self._loaded_keys

    def __setitem__(self, key, value):
        self._value[key] = value

    def __getitem__(self, item):
        if isinstance(self._value, dict):
            return self._value.get(item, None)
        elif isinstance(item, int) and isinstance(self._value, list):
            return self._value[item]
        raise NotImplementedError()

    def __repr__(self):
        return self.to_json(indent=2)

    @property
    def prop(self):
        if self.property_name is None:
            return self.__class__.__name__.lower()
        return self.property_name

    def to_dict(self):
        if isinstance(self._value, _Null):
            return {
                self._name: None
            }

        values = recur_to_dict(self._value)
        if self.union_type == 'dict' and isinstance(values, list):
            _v = {}
            [_v.update(i) for i in values]
            values = _v

        if self._name is None:
            return values

        if self._name == '':
            val = {}
            if self.include_type:
                val['type'] = self.prop
            val[self.prop] = values
            return val

        val = {}
        if self.include_type:
            val['type'] = self.prop
        val[self.prop] = values

        return {
            self._name: val
        }

    def to_json(self, indent=0):
        return json.dumps(self.to_dict(), ensure_ascii=False, indent=indent)

    def parse_schema(self, name, dic: dict):
        """
        parse_schema(k,v) for k,v in properties.items()
        :param name:
        :param dic:
        :return:
        """


class Parent(NotionObject):
    def __init__(self, type=None, page_id=None, workspace: bool = None, database_id=None):
        super().__init__(None, kwargs=dict(type=type, page_id=page_id,
                                           workspace=workspace, database_id=database_id))

    @property
    def type(self):
        return self['type']

    @property
    def is_page(self):
        return self.type == 'page_id'

    @property
    def is_database(self):
        return self.type == 'database_id'

    @property
    def is_workspace(self):  # root
        return self.type == 'workspace'

    @property
    def id(self):
        if self.is_page:
            return self['page_id']
        elif self.is_database:
            return self['database_id']
        return None

    @staticmethod
    def PageParent(page_id):
        return Parent(type='page_id', page_id=page_id)

    @staticmethod
    def DataBaseParent(database_id):
        return Parent(type='database_id', database_id=database_id)

    @staticmethod
    def PageWorkspaceParent():
        return Parent(type='workspace', workspace=True)


class Properties(NotionObject):
    union_type = 'dict'

    def __init__(self, *property: NotionObject):
        super().__init__(None, args=property)

    @staticmethod
    def from_dict(dic):
        return Properties(dic)


class Any(NotionObject):
    def __init__(self, pname: str, property_name=None, union_type='list', include_type=False, args=None, kwargs=None):
        self.property_name = property_name
        self.union_type = union_type
        self.include_type = include_type
        super().__init__(pname, args=args, kwargs=kwargs)


def recur_to_dict(object):
    if isinstance(object, (List, Tuple)):
        return [recur_to_dict(i) for i in object]
    elif isinstance(object, Dict):
        return {k: recur_to_dict(v) for k, v in object.items()}
    elif isinstance(object, NotionObject):
        return object.to_dict()
    return object



from .common import NotionObject


class RichText(NotionObject):
    # property_name = 'rich_text'
    include_type = True

    def __init__(self, kwargs):
        if self.__class__ == RichText:
            self.property_name = 'rich_text'
        super().__init__('', kwargs=kwargs)


class Text(RichText):

    def __init__(self, content, link=None):
        super(Text, self).__init__(kwargs=dict(content=content, link=link))


class Link(RichText):
    """
    Text link objects contain a type key whose value is always "url" and a url key whose value is a web address
    """

    def __init__(self, url=None):
        super(Link, self).__init__(kwargs=dict(url=url))



import json
from typing import Union, overload, Set

from .base import Parser
from ..objects import Page, Database


class MarkdownParser(Parser):
    ext = 'md'

    @overload
    def __init__(self, skipkeys=False, ensure_ascii=True, check_circular=True,
                 allow_nan=True, cls=None, indent=None, separators=None,
                 default=None, sort_keys=False, **kw):
        ...

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def parse(self, doc_info: Union[Page, Database], children=None):
        if isinstance(doc_info, Database):
            return parse_database(doc_info, children)
        elif isinstance(doc_info, Page):
            return parse_page(doc_info)
        else:
            raise NotImplementedError()


def parse_database(d: Database, children: Set[Page]):
    pass


def parse_page(p: Page):
    return json.dumps(p.to_dict(), indent=2, ensure_ascii=False)



import os
from typing import TYPE_CHECKING

from potion import Request
from potion.api import *
from potion.objects import Parent

if TYPE_CHECKING:
    from .page import NotionPage
    from .database import NotionDatabase


class NotionObject:
    def __init__(self, auth: str = None, id: str = None, parent: Parent = None):
        if auth is None:
            auth = os.environ.get('NOTION_SECRET', None)
            assert auth is not None, 'no auth'

        self.id = id

        self.req = Request.from_token(authorization=auth)
        self.auth = auth
        self.object = None

        self._parent = parent
        res = self.req.get(user_bot_retrieve())
        print(res)

    @property
    def parent(self):
        if self._parent.is_page:
            return self.parent_page
        elif self._parent.is_database:
            return self.parent_database
        return None

    def __str__(self):
        return str(self.object)

    @property
    def parent_page(self) -> 'NotionPage':
        from .page import NotionPage
        if self._parent.is_page:
            res = NotionPage(auth=self.auth, id=self._parent.id)
            return res

    @property
    def parent_database(self) -> 'NotionDatabase':
        if self._parent.is_page:
            from .database import NotionDatabase
            res = NotionDatabase(auth=self.auth, id=self._parent.id)
            return res



from potion.urls import url


def database_retrieve(database_id):
    """
    get
    :param database_id:
    :return:
    """
    return url.databases(database_id)


def page_retrieve(page_id):
    """
    get
    :param page_id:
    :return:
    """
    return url.pages(page_id)


def page_property_retrieve(page_id, property_id):
    """
    get
    :param page_id:
    :param property_id:
    :return:
    """
    return url.pages(page_id, property_id)


def block_retrieve(block_id):
    """
    get
    :param block_id:
    :return:
    """
    return url.blocks(block_id)


def block_children_retrieve(block_id, page_size: int = 100, start_cursor=None):
    """
    get
    :param block_id:
    :param page_size:
    :return:
    """
    return url.blocks(block_id, page_size=page_size, start_cursor=start_cursor)


def user_retrieve(user_id):
    """
    get
    :param user_id:
    :return:
    """
    return url.users(user_id)


def users_list():
    """
    get
    :return:
    """
    return url.users()


def user_bot_retrieve():
    """
    get
    :return:
    """
    return url.users('me')



from typing import List

from .common import NotionObject


class Pagination(NotionObject):
    def __init__(self, **kwargs):
        super().__init__(None, kwargs=kwargs)

    @property
    def results(self) -> List[NotionObject]:
        from potion.utils.parser import parse
        return [parse(i) for i in self['results']]

    @property
    def next_cursor(self):
        return self['next_cursor']

    @property
    def has_more(self) -> bool:
        return self['has_more']

    @property
    def type(self) -> str:
        return self['type']

    @property
    def object(self):
        return self['object']



from potion.urls import url


def database_update(database_id):
    """
    patch
    :param database_id:
    :return:
    """
    return url.databases(database_id)


def page_update(page_id):
    """
    patch
    :param page_id:
    :return:
    """
    return url.pages(page_id)


def block_update(block_id):
    """
    patch
    :param block_id:
    :return:
    """
    return url.blocks(block_id)


def block_children_append(block_id):
    """
    patch
    :param block_id:
    :return:
    """
    return url.blocks(block_id, append=True)



import json
from typing import Dict, Union

import requests

from potion.objects import NotionObject
from .parser import parse


def to_json(data):
    if issubclass(data.__class__, NotionObject):
        data = data.to_json()
    elif isinstance(data, dict):
        data = json.dumps(data, ensure_ascii=False)
    else:
        raise NotImplementedError()
    return data


class Request:
    def __init__(self, headers=None):
        if headers is None:
            from potion.const import current_headers
            headers = current_headers
        assert 'Authorization' in headers
        self.headers = headers

    def get(self, url):
        content = json.loads(requests.get(url, headers=self.headers).content)
        return parse(dic=content)

    def post(self, url, data: Union[NotionObject, Dict, str]):
        data = to_json(data)
        content = json.loads(requests.post(url=url, data=data.encode(), headers=self.headers).content)
        return parse(dic=content)

    def patch(self, url, data: Union[NotionObject, Dict, str]):
        data = to_json(data)
        content = json.loads(requests.patch(url=url, data=data.encode(), headers=self.headers).content)
        return parse(dic=content)

    def delete(self, url):
        content = json.loads(requests.delete(url, headers=self.headers).content)
        return parse(dic=content)

    def parse(self):
        pass

    @staticmethod
    def from_token(authorization):
        from potion.const import NotionHeader
        nh = NotionHeader(authorization=authorization)
        return Request(nh.headers)



from typing import Union, List, Dict

from potion.objects import NotionObject
from potion.objects.block import Block
from potion.objects.common import Properties, Parent
from potion.objects.property import Property


class Page(NotionObject):
    """
    For appending blocks
    ```python
    Page(children=[
        block.Todo(rich_text=[], checked=True, color=Color.blue_background),
        block.Heading2(rich_text=[{"type": "text", "text": {"content": "Lacinato kale"}}]),
        {
            "type": "heading_2",
            "heading_2": {
                "rich_text": [{"type": "text", "text": {"content": "Lacinato kale"}}]
            }
        },
    ])
    ```
    For create page
    ```python
    properties = Properties(
        prop.Title('Name',
                   rich_text=[
                       {
                           "text": {
                               "content": "Tuscan Kale"
                           }
                       }
                   ]
                   ),
        prop.Number('Price', 2.5),
        prop.MultiSelect('Args', [
            prop.MultiSelectOption(name='A'),
            prop.MultiSelectOption(name='B'),
        ])
    )

    data = Page(parent=Parent.DataBaseParent('7d39de23304d484ea079a35f816ae68f'),
                properties=properties
                )
    ```
    """
    union_type = 'dict'

    def __init__(self,
                 parent: Union[Parent, Dict] = None,
                 title=None,
                 icon=None,
                 properties: Union[Properties, List[Property]] = None,
                 children: List[Union['Block', Dict]] = None, **kwargs):
        if isinstance(parent, dict):
            parent = Parent(**parent)
        if isinstance(properties, list):
            properties = Properties(*properties)

        super().__init__(None,
                         kwargs=dict(parent=parent,
                                     title=title,
                                     icon=icon,
                                     children=children,
                                     properties=properties,
                                     **kwargs))

        self.load_mark('parent')
        self.load_mark('properties')

    @property
    def string_title(self) -> str:
        for k, v in self['properties'].items():
            if v['id'] == 'title':
                return ''.join([t['plain_text'] for t in v['title']])
        return self.id

    @property
    def parent(self) -> Parent:
        return self['parent']

    @property
    def properties(self) -> Properties:
        return self['properties']

    @property
    def object(self):
        return self['object']

    @property
    def short_id(self):
        return self.id[:4]

    @property
    def id(self):
        return self['id']

    @property
    def created_time(self):
        return self['created_time']

    @property
    def last_edited_time(self):
        return self['last_edited_time']

    @property
    def created_by(self):
        return self['created_by']

    @property
    def last_edited_by(self):
        return self['last_edited_by']

    @property
    def archived(self):
        return self['archived']

    @property
    def url(self):
        return self['url']

    def set_children(self, blocks):
        self['children'] = blocks



from . import block
from . import property as prop
from . import query
from . import rich
from . import schema as sche
from .block import Block
from .common import NotionObject, Parent, Properties, Null
from .database import Database
from .enums import Color, Language, FilterProperty, FilterCondition
from .error import Error
from .page import Page
from .pagination import Pagination
from .query import Filter, Sorts, Search



from potion.urls import url


def database_create():
    """
    post
    :return:
    """
    return url.databases()


def database_query(database_id, ):
    """
    post
    :param database_id:
    :return:
    """
    return url.databases(database_id, query=True)


def page_create():
    """
    post
    :return:
    """
    return url.pages()


def search():
    """
    post
    :return:
    """
    return url.search()



import re

from setuptools import setup, find_packages


def extract_version():
    return re.search(
        r"""__version__ = ['"]([\d.d\-]+)['"]""",
        open('src/potion/__init__.py', 'r', encoding='utf-8').read()).group(1)


if __name__ == '__main__':
    setup(
        name='notion-potion',
        version=extract_version(),
        description='A functional, easy to use Python wrapper of Notion Api.',
        long_description_content_type='text/markdown',
        url='https://github.com/sailist/potion',
        author='sailist',
        author_email='sailist@outlook.com',
        license='Apache License 2.0',
        include_package_data=True,
        install_requires=['dbrecord', 'requests'],
        # https://pypi.org/classifiers/
        classifiers=[
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            "Operating System :: OS Independent",
            "License :: OSI Approved :: GNU General Public License v3 (GPLv3)"
        ],
        keywords='potion notion',
        package_dir={"": "src"},
        packages=find_packages('src'),
        # packages=find_packages('.', exclude=('tests', 'scripts', 'examples')),
        entry_points={},
    )



import json
from typing import Union, overload

from .base import Parser
from ..objects import Page, Database


class HTMLParser(Parser):
    ext = 'html'

    @overload
    def __init__(self, skipkeys=False, ensure_ascii=True, check_circular=True,
                 allow_nan=True, cls=None, indent=None, separators=None,
                 default=None, sort_keys=False, **kw): ...

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def parse(self, doc_info: Union[Page, Database]):
        return json.dumps(doc_info.to_dict(), **self._config)


def parse_database(d: Database):
    pass


def parse_page(p: Page):
    pass



"""
Database objects describe the property schema of a database in Notion.
Pages are the items (or children) in a database.
Page property values must conform to the property objects laid out in the parent database object.

"""

from .common import NotionObject, Parent


class Database(NotionObject):
    union_type = 'dict'

    def __init__(self, parent=None, title=None, icon=None, properties=None, **kwargs):
        if isinstance(parent, dict):
            parent = Parent(**parent)

        super().__init__(None,
                         kwargs=dict(parent=parent,
                                     title=title,
                                     icon=icon,
                                     properties=properties,
                                     **kwargs))

    @property
    def short_id(self):
        return self.id[:4]

    @property
    def string_title(self):
        return ''.join([i['plain_text'] for i in self.title])

    @property
    def parent(self) -> Parent:
        return self['parent']

    @property
    def title(self):
        return self['title']

    @property
    def icon(self):
        return self['icon']

    @property
    def properties(self):
        return self['properties']

    @property
    def object(self) -> str:
        return self['object']

    @property
    def id(self) -> str:
        return self['id']

    @property
    def created_time(self):
        return self['created_time']

    @property
    def created_by(self):
        return self['created_by']

    @property
    def last_edited_by(self):
        return self['last_edited_by']

    @property
    def last_edited_time(self):
        return self['last_edited_time']

    @property
    def url(self) -> str:
        return self['url']

    @property
    def archived(self) -> bool:
        return self['archived']



import os
import sys
import time
from typing import Union

from dbrecord.pdict import PDict
from tqdm import tqdm

from potion import Request, NotionHeader
from potion.api import *
from potion.objects import *


def recursive_parsechild(parent: Union[Page, Block], pid):
    has_more = True
    blocks = []
    start_cursor = None
    while has_more:
        pag = req.get(url=block_children_retrieve(pid, start_cursor=start_cursor))
        if isinstance(pag, Pagination):
            has_more = pag.has_more
            start_cursor = pag.next_cursor
            for block in pag.results:
                if isinstance(block, Block):
                    if block.has_children and block.type not in {'child_page', 'child_database'}:
                        recursive_parsechild(block, block.id)
                blocks.append(block)
        else:
            time.sleep(1)
            continue

    parent.set_children(blocks)

    return parent


def full_backup(req, base_path):
    object_value = PDict(os.path.join(base_path, 'notion-backup.sqlite'))

    has_more = True
    start_cursor = None

    skiped, updated = 0, 0
    while has_more:
        data = Search(query='',
                      sort={
                          "direction": "ascending",
                          "timestamp": "last_edited_time"
                      },
                      page_size=0,
                      start_cursor=start_cursor)
        pag = req.post(url=search(),
                       data=data)

        if isinstance(pag, Pagination):
            has_more = pag.has_more
            start_cursor = pag.next_cursor
            for res in tqdm(pag.results):
                if res.id in object_value:
                    if object_value[res.id]['last_edited_time'] == res.last_edited_time:
                        skiped += 1
                        continue

                if isinstance(res, Page):
                    res = recursive_parsechild(res, res.id)
                elif isinstance(res, Database):
                    pass
                else:
                    raise NotImplementedError()

                object_value[res.id] = res
                updated += 1
        else:
            time.sleep(0.5)
            continue
    object_value.flush()

    return skiped, updated


if __name__ == '__main__':
    args = list(sys.argv[1:])
    if len(args) < 2:
        print('python -m potion.backup {token} {path}')
        exit(1)

    token, base_path, *_ = sys.argv[1:]

    os.makedirs(base_path, exist_ok=True)

    nh = NotionHeader(authorization=token)
    req = Request(nh.headers)

    skiped, updated = full_backup(req, base_path)
    print(f'Notion backup finished. {skiped} skiped, {updated} updated.')
    exit(0)



import os
import os
import re
from collections import defaultdict
from pathlib import Path
from typing import Union, List

from potion.objects import Page, Database


class Parser:
    ext = '.json'

    def __init__(self, **kwargs):
        self._config = kwargs

    def parse(self, doc_info: Union[Page, Database], children=None):
        raise NotImplementedError()


match_invalid = re.compile('/')


def safe_path(title):
    return match_invalid.sub('_', title)


class Node:
    def __init__(self, value, children=None):
        self.value = value
        self.children = set() if children is not None else children


class DocumentRelation:
    def __init__(self, documents: List[Union[Page, Database]], meta=None):
        self.meta = meta
        self.roots = set()
        self.tree = defaultdict(set)
        self.rev_tree = {}
        self.id_doc_map = {}

        for doc in documents:
            if doc.parent.type == 'workspace':
                self.root = doc
            else:
                pid = doc.parent[doc.parent.type]
                self.rev_tree[doc.id] = pid
                self.tree[pid].add(doc.id)

            self.id_doc_map[doc.id] = doc

        mem = set()
        node_index = {}

        for cid, pid in self.rev_tree.items():
            if pid not in self.rev_tree and pid in self.id_doc_map:
                self.roots.add(pid)

    def parse_org(self, parser: Parser, tgt_path: Path):
        for root_id in self.roots:
            self.parse_tree(parser, root_id, tgt_path)

    def parse_tree(self, p: Parser, cid, tgt_path: Path):
        os.makedirs(str(tgt_path), exist_ok=True)

        doc = self.id_doc_map[cid]
        title = f'{doc.string_title}_{doc.short_id}'

        children = self.tree[cid]
        content = p.parse(doc, children)

        tgt_file = tgt_path / safe_path(f'{title}.{p.ext}')
        try:
            tgt_file.write_text(content)
        except:
            print(tgt_file)

        # parse children
        for child in children:
            self.parse_tree(p, child, tgt_path)
